import argparse
import asyncio
import asyncio.exceptions
import contextlib
import datetime
import functools
import importlib
import logging
import os
import sys
import time
import urllib.parse
from typing import Any, Optional

from aio_pika import IncomingMessage
from yarl import URL

from .annotated_task import AnnotatedTask
from .app import Celery
from .task import (
    RetryRequested,
    Task,
)

BANNER = """\
[config]
.> app:         {app}
.> transport:   {conninfo}
.> results:     {results}
.> concurrency: {concurrency}

[queues]
{queues}
"""

logger = logging.getLogger(__name__)


async def _sleep_if_necessary(task: Task) -> None:
    eta: Optional[datetime.datetime] = task.eta
    if eta is None:
        return
    now = datetime.datetime.now().astimezone()
    if eta > now:
        logger.info("[%s] Sleeping until %s...", task.task_id, eta)
        await asyncio.sleep((eta - now).total_seconds())


async def _handle_task_result(
    task: Task,
    annotated_task: AnnotatedTask,
    result: Any,
) -> None:
    if (
        task.app.conf.task_ignore_result
        or annotated_task.ignore_result
        or task._get_task_ignore_result()
    ):
        return
    await task.update_state(state="SUCCESS", meta=result, _finalize=True)


async def on_message_received(message: IncomingMessage, *, app: Celery) -> None:
    timestamp_start = time.monotonic()
    task_id = None
    task_name = None
    task_args = None
    task_kwargs = None
    try:
        async with message.process(ignore_processed=True):
            task = Task.from_raw_message(message, app=app)
            task_id = task.task_id
            task_name = task.task_name
            task_args = task.args
            task_kwargs = task.kwargs
            logger.info("Task %s[%s] received", task.task_name, task.task_id)
            try:
                annotated_task = app._get_annotated_task(task.task_name)
            except KeyError:
                logger.exception(
                    "Received unregistered task of type %r\n"
                    "The message has been ignored and discarded.\n\n"
                    "Did you remember to import the module containing this task?\n"
                    "Or maybe you're using relative imports?\n\n"
                    "Please see\n"
                    "https://docs.celeryq.dev/en/latest/internals/protocol.html\n"
                    "for more information.\n\n"
                    "The full contents of the message body was:\n"
                    "%r (%db)\n\n"
                    "The full contents of the message headers:\n"
                    "%r\n\n"
                    "The delivery info for this task is:\n"
                    "%r",
                    task.task_name,
                    message.body,
                    len(message.body),
                    message.headers,
                    {
                        "consumer_tag": message.consumer_tag,
                        "delivery_tag": message.delivery_tag,
                        "redelivered": message.redelivered,
                        "exchange": message.exchange,
                        "routing_key": message.routing_key,
                    },
                )
                await message.reject()
                return

            await _sleep_if_necessary(task)
            try:
                args = (task, *task.args) if annotated_task.bind else task.args
                async with asyncio.timeout(task.task_soft_time_limit):
                    result = await annotated_task.fn(*args, **task.kwargs)
            except RetryRequested as exc:
                if (
                    annotated_task.max_retries is not None
                    and task.get_already_happened_retries() > annotated_task.max_retries
                ):
                    raise RuntimeError(
                        f"Max retries exceeded {annotated_task.max_retries=}",
                    ) from exc
                await app._publish(
                    exc.message,
                    routing_key=(message.routing_key or app.conf.task_default_queue),
                )
            else:
                await _handle_task_result(
                    task=task,
                    annotated_task=annotated_task,
                    result=result,
                )
                if task.chain:
                    next_message, next_routing_key = task._build_next_task_message(
                        result=result,
                    )
                    await app._publish(next_message, routing_key=next_routing_key)
                logger.info(
                    "Task %s[%s] succeeded in %.6fs: %r",
                    task.task_name,
                    task.task_id,
                    time.monotonic() - timestamp_start,
                    result,
                )
    except Exception:
        logger.exception(
            "[%s] Unexpected error happened: [%s] args=%r, kwargs=%r",
            task_id,
            task_name,
            task_args,
            task_kwargs,
        )


@contextlib.contextmanager
def _cwd_in_path():
    """Context adding the current working directory to sys.path."""
    cwd = os.getcwd()
    if cwd in sys.path:
        yield
    else:
        sys.path.insert(0, cwd)
        try:
            yield cwd
        finally:
            with contextlib.suppress(ValueError):
                sys.path.remove(cwd)


def _find_app_instance(locator: str) -> Celery:
    with _cwd_in_path():
        module_name, app_name = locator.rsplit(":", 1)
        module = importlib.import_module(module_name)
        return getattr(module, app_name)


def _print_intro(concurrency, queue_names, task_names, app):
    q = ".> " + "\n.> ".join(queue_names)
    t = "  . " + "\n  . ".join(task_names)

    def _normalize(u):
        if not u:
            return "disabled://"
        parts = urllib.parse.urlparse(u)
        if not parts.password:
            return u
        return str(URL(u).with_password("******"))

    banner = BANNER.format(
        app=repr(app),
        conninfo=_normalize(app.conf.broker_url),
        results=_normalize(app.conf.result_backend),
        concurrency=f"{concurrency} (asyncio)",
        queues=q,
    ).splitlines()
    print(
        "╔═══════════════════════╗\n"
        "║ AsyncIO Celery Worker ║\n"
        "╚═══════════════════════╝\n",
    )
    print("\n".join(banner) + "\n")
    print(f"[tasks]\n{t}\n")


async def run(args: argparse.Namespace) -> None:
    logging.basicConfig(
        level=logging.getLevelName(args.loglevel),
        format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
    )
    logging.getLogger("aio_celery").setLevel(level=logging.getLevelName(args.loglevel))
    app = _find_app_instance(args.app)
    if not args.queues:
        queue_names = [app.conf.task_default_queue]
    else:
        queue_names = [q.strip() for q in args.queues.split(",")]

    _print_intro(
        args.concurrency,
        queue_names,
        app.list_registered_task_names(),
        app,
    )

    async with app.setup():
        await app.rabbitmq_channel.set_qos(prefetch_count=args.concurrency)
        queues = [
            await app.rabbitmq_channel.declare_queue(
                name=name,
                durable=True,
                arguments={"x-max-priority": app.conf.task_queue_max_priority},
            )
            for name in queue_names
        ]
        for queue in queues:
            await queue.consume(
                functools.partial(on_message_received, app=app),
                timeout=None,  # this should be soft time limit
            )
        logger.info("Waiting for messages. To exit press CTRL+C")
        await asyncio.Future()



import copy
import datetime
import json
import logging
import uuid
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from aio_pika import IncomingMessage, Message

from .amqp import create_task_message
from .app import Celery

logger = logging.getLogger(__name__)


@dataclass
class Task:
    message: IncomingMessage
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]
    options: Dict[str, Any]
    chain: List[Dict[str, Any]]

    app: Celery

    @property
    def task_id(self) -> str:
        return str(self.message.headers["id"])

    @property
    def task_name(self) -> str:
        return str(self.message.headers["task"])

    @property
    def eta(self) -> Optional[datetime.datetime]:
        eta = self.message.headers["eta"]
        if eta is None:
            return None
        return datetime.datetime.fromisoformat(eta)

    @classmethod
    def from_raw_message(
        cls,
        message: IncomingMessage,
        *,
        app: Celery,
    ) -> "Task":
        args, kwargs, options = json.loads(message.body)
        callbacks, errbacks, chain, chord = (
            options["callbacks"],
            options["errbacks"],
            options["chain"],
            options["chord"],
        )
        assert callbacks is None
        assert errbacks is None
        assert chord is None
        return cls(
            message=message,
            args=args,
            kwargs=kwargs,
            options=options,
            chain=chain,
            app=app,
        )

    def get_already_happened_retries(self) -> int:
        retries = int(self.message.headers["retries"])
        assert retries >= 0
        return retries

    async def update_state(
        self,
        *,
        state: str,
        meta: Dict[str, Any],
        _finalize: bool = False,
    ) -> None:
        result_backend = self.app.result_backend
        if result_backend is None:
            logger.debug("Result backend has not been enabled")
            return
        if _finalize:
            # do not update if result already in redis
            pass
        payload = {
            "status": state,
            "result": meta,
            "traceback": None,
            "children": [],
            "date_done": None,
            "task_id": self.task_id,
        }
        if _finalize:
            payload["date_done"] = (
                datetime.datetime.utcnow().isoformat()  # noqa: DTZ003
            )
        if self.message.headers["group"] is not None:
            payload["group_id"] = self.message.headers["group"]
        if self.message.headers["parent_id"] is not None:
            payload["parent_id"] = self.message.headers["parent_id"]
        await result_backend.set(
            f"celery-task-meta-{self.task_id}",
            json.dumps(payload).encode(),
            ex=self.app.conf.result_expires,
        )

    def _build_next_task_message(self, result: Any) -> Tuple[Message, str]:
        assert self.chain
        new_chain = copy.deepcopy(self.chain)
        first = new_chain.pop()
        opts = first.get("options", {})
        new_task_id: str = opts.get("task_id") or str(uuid.uuid4())
        new_priority: int = opts.get("priority") or self.app.conf.task_default_priority
        routing_key: str = opts.get("queue") or self.app.conf.task_default_queue
        reply_to: str = opts.get("reply_to") or ""
        return (
            create_task_message(
                task_id=new_task_id,
                task_name=first["task"],
                args=(result, *first["args"]),
                kwargs=first["kwargs"],
                priority=new_priority,
                parent_id=self.task_id,
                chain=new_chain,
                reply_to=reply_to,
            ),
            routing_key,
        )

    def _build_retry_message(
        self,
        *,
        countdown: Optional[datetime.timedelta],
    ) -> Message:
        message = copy.copy(self.message)
        if countdown is None:
            eta = None
        else:
            eta = (datetime.datetime.now().astimezone() + countdown).isoformat()
        message.headers.update(
            eta=eta,
            retries=self.message.headers["retries"] + 1,
        )
        return message

    def retry(self, *, countdown: Optional[float] = None) -> None:
        raise RetryRequested(
            message=self._build_retry_message(
                countdown=(
                    datetime.timedelta(seconds=countdown)
                    if countdown is not None
                    else None
                ),
            ),
        )

    @property
    def task_soft_time_limit(self) -> Optional[int]:
        soft, _hard = self.message.headers["timelimit"]
        return soft or self.app.conf.task_soft_time_limit

    def _get_task_ignore_result(self) -> bool:
        return bool(self.message.headers["ignore_result"])


class RetryRequested(Exception):
    def __init__(self, *, message: Message) -> None:
        self.message = message
        super().__init__()



import contextlib
import logging
import sys
import uuid
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

import aio_pika

if TYPE_CHECKING:
    import redis.asyncio

from .amqp import create_task_message
from .annotated_task import AnnotatedTask
from .backend import create_redis_connection_pool
from .config import DefaultConfig
from .result import AsyncResult as _AsyncResult

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _CompleteTaskResources:
    context: Any
    redis_client_celery: Any


class Celery:
    def __init__(self) -> None:
        self.conf = DefaultConfig()
        self._tasks_registry: Dict[str, AnnotatedTask] = {}
        self._app_context: Any = None
        self._result_backend_connection_pool: Optional[
            "redis.asyncio.BlockingConnectionPool"
        ] = None
        self.rabbitmq_channel: Optional[aio_pika.RobustChannel] = None
        self._setup_app_context: Callable[
            [],
            contextlib.AbstractAsyncContextManager,
        ] = _setup_nothing

    def define_app_context(
        self,
        fn: Callable[[], contextlib.AbstractAsyncContextManager],
    ) -> None:
        self._setup_app_context = fn

    @property
    def context(self) -> Any:
        return self._app_context

    @property
    def result_backend(self) -> Optional["redis.asyncio.Redis"]:
        if self._result_backend_connection_pool is None:
            return None

        from redis.asyncio import Redis

        return Redis(connection_pool=self._result_backend_connection_pool)

    @contextlib.asynccontextmanager
    async def setup(self) -> AsyncIterator[None]:
        connection = await aio_pika.connect_robust(self.conf.broker_url)
        async with connection, connection.channel() as channel:
            self.rabbitmq_channel = channel
            try:
                if self.conf.result_backend is not None:
                    self._result_backend_connection_pool = create_redis_connection_pool(
                        url=self.conf.result_backend,
                        pool_size=self.conf.result_backend_connection_pool_size,
                    )
                async with self._setup_app_context() as context:
                    self._app_context = context
                    try:
                        yield
                    finally:
                        logger.warning("Shutting down application.")
            finally:
                if self._result_backend_connection_pool is not None:
                    await self._result_backend_connection_pool.disconnect()

    def task(self, *args, **opts):
        """Decorator to create a task class out of any callable."""

        def create_annotated_task(
            *,
            bind: bool = False,
            name: Optional[str] = None,
            ignore_result: Optional[bool] = None,
            max_retries: Optional[int] = None,
        ):
            def decorator(fn):
                if name is None:
                    task_name = _gen_task_name(fn.__name__, fn.__module__)
                else:
                    task_name = name
                annotated_task = AnnotatedTask(
                    fn=fn,
                    bind=bind,
                    ignore_result=ignore_result,
                    max_retries=max_retries,
                    task_name=task_name,
                    app=self,
                )
                self._tasks_registry[task_name] = annotated_task
                return annotated_task

            return decorator

        if len(args) == 1:
            if callable(args[0]):
                return create_annotated_task(**opts)(*args)
            raise TypeError("argument 1 to @task() must be a callable")
        if args:
            raise TypeError(
                f"@task() takes exactly 1 argument ({len(args) + len(opts)} given)",
            )
        return create_annotated_task(**opts)

    def _get_annotated_task(self, task_name: str) -> AnnotatedTask:
        return self._tasks_registry[task_name]

    def list_registered_task_names(self) -> List[str]:
        return sorted(self._tasks_registry)

    def AsyncResult(self, task_id: str) -> _AsyncResult:
        return _AsyncResult(task_id, app=self)

    async def send_task(  # noqa: PLR0913
        self,
        name: str,
        *,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        countdown: Optional[int] = None,
        task_id: Optional[str] = None,
        priority: Optional[int] = None,
        queue: Optional[str] = None,
    ) -> _AsyncResult:
        task_id = task_id or str(uuid.uuid4())
        routing_key = queue or self.conf.task_default_queue
        logger.info("Sending task %s[%s] to queue %r ...", name, task_id, routing_key)
        await self._publish(
            create_task_message(
                task_id=task_id,
                task_name=name,
                args=args,
                kwargs=kwargs,
                priority=priority,
                countdown=countdown,
            ),
            routing_key=routing_key,
        )
        return self.AsyncResult(task_id)

    async def _publish(self, message: aio_pika.Message, routing_key: str) -> None:
        assert self.rabbitmq_channel is not None
        await self.rabbitmq_channel.default_exchange.publish(
            message,
            routing_key=routing_key,
            timeout=60,
        )


@contextlib.asynccontextmanager
async def _setup_nothing() -> AsyncIterator[None]:
    yield None


def _gen_task_name(name: str, module_name: str) -> str:
    """Generate task name from name/module pair."""
    module_name = module_name or "__main__"
    module = sys.modules[module_name]
    if module is not None:
        module_name = module.__name__
    return ".".join(p for p in (module_name, name) if p)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import redis.asyncio


def create_redis_connection_pool(
    *,
    url: str,
    pool_size: int,
) -> "redis.asyncio.BlockingConnectionPool":
    from redis.asyncio import BlockingConnectionPool
    from redis.asyncio.retry import Retry
    from redis.backoff import EqualJitterBackoff

    try:
        from redis.backoff import DEFAULT_BASE, DEFAULT_CAP
    except ImportError:
        # Maximum backoff between each retry in seconds
        DEFAULT_CAP = 0.512
        # Minimum backoff between each retry in seconds
        DEFAULT_BASE = 0.008

    assert url.startswith("redis://")

    return BlockingConnectionPool.from_url(
        url=url,
        max_connections=pool_size,
        timeout=None,
        socket_timeout=5,
        retry_on_timeout=True,
        retry=Retry(
            backoff=EqualJitterBackoff(cap=DEFAULT_CAP, base=DEFAULT_BASE),
            retries=5,
        ),
    )



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Optional

if TYPE_CHECKING:
    from .app import Celery
from .result import AsyncResult


@dataclass(frozen=True)
class AnnotatedTask:
    fn: Callable[..., Any]
    bind: bool
    ignore_result: Optional[bool]
    max_retries: Optional[int]
    app: "Celery"
    task_name: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if self.bind:
            return self.fn(None, *args, **kwargs)
        return self.fn(*args, **kwargs)

    async def apply_async(self, **options: Any) -> AsyncResult:
        options.setdefault("name", self.task_name)
        return await self.app.send_task(**options)

    async def delay(self, *args: Any, **kwargs: Any) -> AsyncResult:
        return await self.apply_async(args=args, kwargs=kwargs)
